
#include <qinputdialog.h>
#include "brainApp.h"
#include "extselection.h"

#define NUM_PAINTING_LABELS 32
static unsigned char rgbt[] = { 255, 255, 255,   // white
                                255,   0,   0,   // red
                                  0, 255,   0,   // green
                                  0,   0, 255,   // blue
                                // from http://cloford.com/resources/colours/500col.htm
                                255, 181, 197,   // pink 1
                                205, 105, 201,   // orchid 3
                                132, 112, 255,   // lightslateblue
                                 78, 238, 148,   // seagreen 2
                                255, 255,   0,   // yellow 1
                                255, 185,  15,   // darkgoldenrod 1
                                255, 231, 186,   // wheat 1
                                 56, 142, 142,   // sgi teal
                                197, 193, 170,   // sgi brightgray
                                139,  26,  26,   // firebrick 4
                                139, 137, 137,   // snow 4
                                244, 164,  96,   // sandybrown
                                255, 215,   0,   // gold 1
                                205, 205,   0,   // yellow 3
                                118, 238,   0,   // chartreuse 2
                                189, 252, 201,   // mint
                                 72, 209, 204,   // mediumturquoise
                                255,   0, 255,   // magenta
                                205, 181, 205,   // thistle 1
                                145,  44, 238,   // purple 2
                                 72, 118, 255,   // royalblue 1
                                  0, 255, 127,   // springgreen
                                107, 142,  35,   // olivedrab
                                237, 145,  33,   // carrot
                                158, 158, 158,   // gray 62
                                 26,  26,  26,   // gray 10
                                238,  59,  59,   // brown 2
                                124, 205, 124    // palegreen 3
                              };

SoSeparator* bic_graphics_file_to_iv( const char* filename );

SoTextureCoordinate2* vertexColumn_to_texture( vertexColumn texture,
                                               bool invert) {

  SoTextureCoordinate2 *texCoord = new SoTextureCoordinate2();
  vertexColumn::iterator it;
  int i = 0;
  for (it = texture.begin(); it != texture.end(); it++) {
    if (invert)
      *it *= -1;
    texCoord->point.set1Value(i, SbVec2f(0.5, *it));
    i++;
  }
  return texCoord;
}

/*! This class encapsulates all of the elements important to the application,
 *  such as the various widgets and rendering nodes. It also takes care of 
 *  the initial argument parsing.
 */
brainApp::brainApp(int argc, char **argv, bool colourBar ) {

  // initialise the two main windows
  qapp = new QApplication( argc, argv);

  // load the initialisation file
  environment = new brainEnvironment();

  // set the main window
  mainWindow = new brainWidget();

  // create the auxilliary dialogs
  vstatsWindow = new vertstatsForm();
  vinfoWindow = new vertexInfo();

  // place all this in a grid layout
  QGridLayout *layout = new QGridLayout(mainWindow, 3, 2 );

  // add a menu bar
  QMenuBar *menu = new QMenuBar(mainWindow);
  // and a file menu
  QPopupMenu *fileMenu = new QPopupMenu( mainWindow );
 
  // and a tools menu
  QPopupMenu *toolsMenu = new QPopupMenu( mainWindow );

  // and a menu to handle the clip pane
  QPopupMenu *clipMenu = new QPopupMenu( mainWindow );

  // and a menu for brain painting
  QPopupMenu *paintMenu = new QPopupMenu( mainWindow );

  viewWindow = new QWidget( mainWindow );
  // set the size for the viewing window
  viewWindow->setMinimumHeight(400);
  viewWindow->setMinimumWidth(400);

  // texture view and info bar
  QVBox *textureBox = new QVBox( mainWindow );
  textureBox->setMaximumWidth( 130 );
  topValue = new QLineEdit( textureBox );
  textureWindow = new QWidget( textureBox );
  textureWindow->setMinimumWidth(130);
  lowValue = new QLineEdit( textureBox );
  textureList = new QComboBox( textureBox );
  textureList->insertStringList( *environment->textureFiles );

  //qapp->setMainWidget( viewWindow );

  SoQt::init( viewWindow );

  // begin working with the actual nodes
  objViewer = new SoQtExaminerViewer( viewWindow );

  textureViewer = new SoQtRenderArea( textureWindow );

  // initialise the accessory widgets and classes
  infoWin = new pointInfoWindow( mainWindow );
  picker = new pointPicker();

  // place all of the widgets into their place
  layout->addMultiCellWidget( menu, 0, 0, 0, 1 );
  layout->addMultiCellWidget( infoWin, 2, 2, 0, 1 );
  layout->addWidget( viewWindow, 1, 1 );
  layout->addWidget( textureBox, 1, 0 );
  layout->setRowStretch(1, 10);
  layout->setColStretch(1, 10);

  cout << "before signals" << endl;
  
  // the screen capture thingy
  capturer = new screenCapture(objViewer);

  // the options dialog
  brainoptions = new options(environment);

  // create the R interface without actually starting any processes
  Rpipe = new Rinterface(false);

  // the R plotter
  plotter = new RPlot( Rpipe );

  // the anatomical connectivity widget
  anatconTracer = new anatcon( Rpipe );

  // insert the menu items
  fileMenu->insertItem( "New &File", this , SLOT(newFile()), CTRL+Key_O );
  fileMenu->insertItem( "&Save Inventor Scene-graph", this, 
			SLOT(saveSceneGraph()), CTRL+Key_S );
  fileMenu->insertSeparator();
  fileMenu->insertItem( "Load vertex painting", this, SLOT(loadPainting()));
  fileMenu->insertItem(" Save vertex painting", this, SLOT(savePainting()));
  fileMenu->insertSeparator();
  fileMenu->insertItem( "&Quit", this, SLOT(quit()), CTRL+Key_Q );

  toolsMenu->insertItem( "Vertex Information", this,
                         SLOT( vinfoWindowShow()), CTRL+Key_V );

  toolsMenu->insertItem("Vertex Plotting", this,
			SLOT(plotWindowShow()), CTRL+Key_P);

  // show the anatomical connectivity tool only if the user has the
  // development environment enabled
  if (environment->developmentStatus > 0) {
    toolsMenu->insertItem("Anatomical Connectivity", this,
			  SLOT(anatconWindowShow()), CTRL+Key_A);
  }

  // Possible bug: screen capture crashes the X server on scarus ...
  toolsMenu->insertItem( "Screen Capture", capturer,
			 SLOT( show()), CTRL+Key_S );
  toolsMenu->insertItem( "Background Colour", this,
                         SLOT( setBGColourDialog()) );
  toolsMenu->insertItem( "Options", brainoptions,
                         SLOT( show()) );

  clipMenu->insertItem("Toggle Clip Pane", this, SLOT(toggleClipPane()));
  clipMenu->insertItem("Flip Clip Pane", this, SLOT(flipClipPane()));

  paintMenu->insertItem("Toggle Painting", this, SLOT(togglePainting()));
  paintMenu->insertSeparator();
  paintMenu->insertItem("Set Paint Label", this, SLOT(setPaintLabel()));
  paintMenu->insertItem("Erase painting", this, SLOT(resetPaint()));

  menu->insertItem( "&File", fileMenu );
  menu->insertItem( "&Tools", toolsMenu );
  menu->insertItem( "&Clipping", clipMenu );

  menu->insertItem( "&Painting", paintMenu );

  cout << "initialising" << endl;

 
  // create the various Inventor nodes
  rootObj = new SoSeparator;
  //rootObj->renderCaching = SoSeparator::OFF;
  textureInfo = new SoSeparator;
  rootObj->ref();
  //rootMinc->ref();
  this->textureInfo->ref();

  geometry = new SoSeparator;
  geometry->setName("geometry");
  tBind = new SoTextureCoordinateBinding;
  tBind->value.setValue(SoTextureCoordinateBinding::PER_VERTEX_INDEXED);

  // show everything filled by default
  geomDrawStyle = new SoDrawStyle;
  geomDrawStyle->style = SoDrawStyleElement::FILLED;


  mouseEventCB = new SoEventCallback;

  camera = new SoPerspectiveCamera;
  camera2 = new SoPerspectiveCamera;

  textureSwitch = new SoSwitch;
  // name it so it can be easily found
  textureSwitch->setName("TEXTURESWITCH");

  //materialSwitch = new SoSwitch;
  //colour = new SoBaseColor;

  // share cameras for now

  // initialise the various texture thingies
  vertstats = new mniVertstatsFile();
  texture = new SoTexture2;

  // bug that has cropped up on scarus: if texture is added to the textureInfo pane then
  // it is no longer rendered on the object.
  if (colourBar == true) {
    textureInfo->addChild(texture);
  }
  haveTexture = false;

  originalTexCoord = new SoTextureCoordinate2;
  texCoord = new SoTextureCoordinate2;
  texCoord->setName("textureCoord");

  // first set up the texture transform - this will apply to both the 
  // colourbar as well as the actual brain
  textureTrans = new SoTexture2Transform;
  textureTrans->scaleFactor.setValue(SbVec2f(1,1));
  textureTrans->setName("TextureTransform");

  //   see if texture quality makes any difference
  textureComplexity = new SoComplexity;
  textureComplexity->textureQuality = 1;
  //rootObj->addChild(textureComplexity);

  // material nodes
  material = new SoMaterial;
  //materialBinding = new SoMaterialBinding;
  //materialBinding->value = SoMaterialBinding::PER_VERTEX_INDEXED;

  // painting stuff
  paintingSwitch = new SoSwitch;
  paintingTexture = new SoTexture2;
  paintingTexture->image.setValue(SbVec2s(1,NUM_PAINTING_LABELS), 3, NULL);
  SbVec2s tmpVec;
  int tmpInt;
  unsigned char *t = paintingTexture->image.startEditing(tmpVec, tmpInt);

  for( int ii = 0; ii < 3*NUM_PAINTING_LABELS; ii++ ) {
    t[ii] = rgbt[ii];
  }
  paintingTexture->image.finishEditing();
  paintingTexture->wrapT = SoTexture2::CLAMP;
  paintingTexture->wrapS = SoTexture2::CLAMP;

  paintingTexCoord = new SoTextureCoordinate2;

  paintingTexTrans = new SoTexture2Transform;
  paintingTexTrans->scaleFactor = SbVec2f(1, 1.0/NUM_PAINTING_LABELS);
  paintingTexBinding = new SoTextureCoordinateBinding;
  paintingTexBinding->value.setValue
    (SoTextureCoordinateBinding::PER_VERTEX_INDEXED);
  paintingTexComplexity = new SoComplexity;
  paintingTexComplexity->textureQuality = 0.01;

  paintBrain = false;
  paintNumber = 1;

  // the clip plane
  clipper = new SoClipPlaneManip;
  clipSwitch = new SoSwitch;

  // sphere to indicate node location
  locationSphereSwitch = new SoSwitch;
  locationSphere = new SoSphere;
  // locationSphere->radius.setValue(3.0);
  locationSphere->radius.setValue(0.5);   // use a smaller value (Claude).

  locationSphereTranslation = new SoTranslation;
  locationSphereMaterial = new SoMaterial;
  locationSphereMaterial->diffuseColor.setValue(1.0, 0.0, 0.0);
  // initialise the colourbar window and info
  initColourbar();


  // connect signals and slots
  connect(picker, SIGNAL(coordinatePicked(float,float,float,int)),
	  infoWin->coordinates, SLOT(setCoordinates(float, float, 
							  float, int)));
  connect(picker, SIGNAL(coordinatePicked(float,float,float,int)),
	  this, SLOT(vinfoPopulate(float, float, float,int)));
  connect(picker, SIGNAL(coordinatePicked(float,float,float,int)),
	  this, SLOT(plotFromPickedPoint(float,float,float,int)));

  connect(picker, SIGNAL(coordinatesPicked(vector<int>)),
	  this, SLOT(paintVertex(vector<int>)));

  connect(picker, SIGNAL(coordinatePicked(float,float,float,int)),
	  this, SLOT(traceFromPickedPoint(float,float,float,int)));

  connect(infoWin->coordinates, SIGNAL(newIndex( int )),
          this, SLOT(setLocationSphere( int )));

  connect(picker, SIGNAL(texturePicked(float,float)),
	  infoWin->textureInfo, SLOT(setTextureData(float,float)));
  connect(topValue, SIGNAL(returnPressed()),
	  this, SLOT(textureRangeChanged()) );
  connect(lowValue, SIGNAL(returnPressed()),
	  this, SLOT(textureRangeChanged()) );
  connect(textureList, SIGNAL(activated(int)),
	  this, SLOT(textureSelected(int)));
  connect(vinfoWindow->columnsList, SIGNAL(selectionChanged()),
  	  this, SLOT(vinfoHeaderSelected()));
  connect(anatconTracer, SIGNAL(newDataFile(QString)),
	  this, SLOT(loadNewVertstatsFile(QString)));
  connect(qapp, SIGNAL(lastWindowClosed()),
	  this, SLOT(quit()));
  connect(this, SIGNAL(newObjFile(QString)),
          anatconTracer, SLOT(setCortex(QString)));
  connect(anatconTracer, SIGNAL(getPaintedVertices(int)),
          this, SLOT(sendPaintedVertices(int)));
  connect(this, SIGNAL(paintedVertices(vector<int>)),
          anatconTracer, SLOT(receivePaintedVertices(vector<int>)));

  connect(mainWindow, SIGNAL(newFilenameDropped(QString)),
	  this, SLOT(receivedNewFilename(QString)));

  connect(brainoptions, SIGNAL(newDrawStyle(const QString&)),
	  this, SLOT(newDrawStyle(const QString&)));

  cout << "after signals" << endl;
  
}

void brainApp::setPaintLabel() {
  paintNumber = QInputDialog::getInteger("Paint Label", "New paint label");
}
           
void brainApp::newDrawStyle(const QString &newStyle) {

  cout << "New Style: " << newStyle << endl;


  if (newStyle == "Filled") 
    geomDrawStyle->style = SoDrawStyleElement::FILLED;  
  else if (newStyle == "Lines")
    geomDrawStyle->style = SoDrawStyleElement::LINES;  
  else if (newStyle == "Points")
    geomDrawStyle->style = SoDrawStyleElement::POINTS;  
}

// the old way of painting - should either be removed or reintroduced as a
// a way to paint individual vertices.
void brainApp::paintVertex(vector<int> indices) {
  //cout << "painting vertex: " <<  << endl;
  /*
  if (paintBrain) {
    vector<int>::iterator it = indices.begin();
    
    SbVec2f *vertValues = paintingTexCoord->point.startEditing();
    while( it != indices.end() ) {
      vertValues[*it].setValue(0, paintNumber);
      it++;
    }
    paintingTexCoord->point.finishEditing();
    //sendPaintedVertices(paintNumber);
  }
  */
}


void brainApp::loadPainting() {
  vertexColumn data;
  mniVertstatsFile paintFile;
  
  QString fn = QFileDialog::getOpenFileName
    ( QString::null, 
      "Paintings (*.txt *.vertstats)",
      mainWindow );

  // maybe the user should be warned that unsaved data is about to go
  // poof?
  if ( !fn.isEmpty() ){
    paintFile.loadFile((char *) fn.latin1());
    data = paintFile.getDataColumn(0);

    int numValues = paintingTexCoord->point.getNum();
    SbVec2f *vertValues = paintingTexCoord->point.startEditing();

    for (int i=0; i < numValues; i++) {
      vertValues[i].setValue(0.5, data[i] );
    }
    paintingTexCoord->point.finishEditing();

    paintFile.close();
  }
  else {
    cerr << "No file loaded" << endl;
  }

  
}

void brainApp::savePainting() {
  vertexColumn data;
  mniVertstatsFile paintFile;

  QFileDialog* fd = new QFileDialog( (QWidget *) NULL, "Save Painting as:", 
				     true );
  fd->setMode( QFileDialog::AnyFile );

  QString fileName;
  if ( fd->exec() == QDialog::Accepted ){
    fileName = fd->selectedFile();

    int numValues = paintingTexCoord->point.getNum();
    SbVec2f *vertValues = paintingTexCoord->point.startEditing();

    for (int i=0; i < numValues; i++) {
      data.push_back(vertValues[i].getValue()[1]);
    }
    paintingTexCoord->point.finishEditing();

    paintFile.putDataColumn(data, "Painting");
    paintFile.writeFile((char *) fileName.latin1(), true);
  }

}

void brainApp::togglePainting() {
  if (paintBrain == true) {
    paintBrain = false;
    paintingSwitch->whichChild = SO_SWITCH_NONE;
    objViewer->setSceneGraph(rootObj);
  }
  else {
    paintBrain = true;
    paintingSwitch->whichChild = SO_SWITCH_ALL;
    objViewer->setSceneGraph(paintRoot);
  }
}

void brainApp::setLocationSphere(int index) {
  //locationSphereSwitch = SO_SWITCH_ALL;
  //cout << "Index: " << index << " " << geomCoordinates->point[index][1] << endl;

  if (environment->showPickedVertex == true) {
    locationSphereSwitch->whichChild = SO_SWITCH_ALL;
    locationSphereTranslation->translation.setValue(geomCoordinates->point[index]);
  }
  else {
    locationSphereSwitch->whichChild = SO_SWITCH_NONE;
  }
}

void brainApp::resetPaint() {
  for (int i=0; i < geomCoordinates->point.getNum(); i++) {
    paintingTexCoord->point.set1Value(i, SbVec2f(0, 0));
  }
}


void brainApp::setTextureImage(  QString imageFilename ) {
  texture->filename.setValue(imageFilename.latin1());
    
  texture->wrapT = SoTexture2::CLAMP;
  texture->wrapS = SoTexture2::CLAMP;
}

void brainApp::setTextureFile( QString texFilename ) {
  mainWindow->setCaption(texFilename);
  vertstats->loadFile( (char *) texFilename.ascii() );
}

/*! Replaces or creates the texture coordinates
 */
void brainApp::setTextureCoordinates( bool invert,
                                      int intColumn,
                                      QString stringColumn ) {

  /* a bit of ugly checking to see which column to use. The ugliness
     comes from the fact that it can be specified either as a string
     or as an int */

  vertexColumn data;
  vertexColumn::iterator it;

  if (stringColumn.length() == 0 && intColumn == -1) {
    cerr << "Gack! Something went horribly wrong in specifying which column of the vertstats file to use!! Must die hideous death. " << endl;
  }
  else if (stringColumn.length() == 0) {
    data = vertstats->getDataColumn(intColumn);
  }
  else {
    data = vertstats->getDataColumn(stringColumn.latin1());
  }

  // get the range of the data column that was selected
  float high = *data.begin();
  float low = high;
  for (it = data.begin(); it < data.end(); it++) {
    if ( *it > high )
      high = *it;
    else if ( *it < low )
      low = *it;
  }

  textureMin = low;
  textureMax = high;

  int id = textureSwitch->findChild( texCoord );
  originalTexCoord = vertexColumn_to_texture(data, invert);
  texCoord = (SoTextureCoordinate2 *) originalTexCoord->copy();

  cout << "ID: " << id << endl;

  if (id != -1) {
    textureSwitch->replaceChild( id, texCoord );
  }
  else {
    haveTexture = true;
    textureSwitch->whichChild = SO_SWITCH_ALL;
  }

  if (environment->updateRangeOnTextureChange)
    setTextureRange(low, high);

}

void brainApp::setTextureRange( float min, float max ) {
  textureTrans->scaleFactor.setValue(SbVec2f(1,1));
  int numValues = texCoord->point.getNum();
  
  SbVec2f *vertValues = texCoord->point.startEditing();
  SbVec2f *realValues = originalTexCoord->point.startEditing();

  for (int i=0; i < numValues; i++) {
    const float *curVal = realValues[i].getValue();
    float new_val = ( curVal[1] - min ) / ( max - min );
    if( new_val < 0.0 ) {
      new_val = 0.0;
    } else if( new_val > 1.0 ) {
      new_val = 1.0;
    }
    vertValues[i].setValue(0.5, new_val );
  }
  texCoord->point.finishEditing();
  originalTexCoord->point.finishEditing();
  currentMin = min;
  currentMax = max;
  
}


void brainApp::loadNewVertstatsFile( QString filename) {
  mainWindow->setCaption(filename);

  vertstats->close();
  setTextureFile( (char *) filename.ascii() );
  setTextureCoordinates(false, 0, "");
  updateRangeWidgets();
  textureSwitch->whichChild = SO_SWITCH_ALL;
  vinfoPopulate();
}

void brainApp::textureSelected(int listIndex) {
  QString newTexture = textureList->currentText();
  newTexture.prepend(*environment->textureDir);
  setTextureImage( (char *) newTexture.latin1());
  textureRangeChanged();
}

void brainApp::textureRangeChanged() {
  // A signal that is invoked whenever the texture range has changed (duh).
  // Sets the the colourbar captions as well as calls the function to 
  // actually change the text in the colour bar.

  char minString[10];
  char maxString[10];
  char threeQString[10];

  // get the new minimum texture range:
  float min;
  if (lowValue->text() == "MIN") 
    min = textureMin;
  else if (lowValue->text() == "MAX")
    min = textureMax;
  else
    min = lowValue->text().toFloat();

  // set the text field
  lowValue->setText(QString("%1").arg(min, 0, 'f', 2));

  // get the nex max texture range:
  float max;
  if (topValue->text() == "MIN") 
    max = textureMin;
  else if (topValue->text() == "MAX")
    max = textureMax;
  else
    max = topValue->text().toFloat();

  // set the text field
  topValue->setText(QString("%1").arg(max, 0, 'f', 2));


  (void) sprintf(minString, "%3.2f", min);
  (void) sprintf(maxString, "%3.2f", max);
  (void) sprintf(threeQString, "%3.2f", float((((max-min)/4)*3)+min));

  lowColourmapValue->string = minString;
  highColourmapValue->string = maxString;
  threeQColourmapValue->string = threeQString;

  setTextureRange( min, max );
}

void brainApp::updateRangeWidgets() {
	// 5A
  // only invoked once upon programme start-up to set the correct
  // initial values. Afterwards textureRangeChanged() is always invoked
  // as a SIGNAL.

  char minString[10];
  char maxString[10];
  (void) sprintf(minString, "%3.2f", currentMin);
  (void) sprintf(maxString, "%3.2f", currentMax);

  lowValue->setText( QString(minString) );
  topValue->setText( QString(maxString) );

  // call this in order to set the captions to the colourbar.
  textureRangeChanged();
}

/*! Replaces or creates the geometry node
 */
void brainApp::setOBJFile(  QString objFilename ) {
  cout << "Opening obj file: " << objFilename << endl;

  int sceneIndex = rootObj->findChild( geometry);
  geometry = bic_graphics_file_to_iv( objFilename.latin1() );
  if (sceneIndex != -1) 
    rootObj->replaceChild( sceneIndex, geometry );
  
  SoSeparator *tmp = new SoSeparator;
  tmp = (SoSeparator *)geometry->getChild(0);
  for (int i=0; i < tmp->getNumChildren(); i++) {
    if ( tmp->getChild(i)->getTypeId() == SoMaterialBinding::getClassTypeId()) {
      materialBinding = (SoMaterialBinding *) tmp->getChild(i);
    }
    else if (tmp->getChild(i)->getTypeId() == SoBaseColor::getClassTypeId()) {
      colour = (SoBaseColor *) tmp->getChild(i);
    }
    else if (tmp->getChild(i)->getTypeId() == SoCoordinate3::getClassTypeId()) {
      geomCoordinates = (SoCoordinate3 *) tmp->getChild(i);
    }
    else { 
      cout << "nothing in: " << i << " name: " << tmp->getChild(i)->getTypeId().getName() <<  endl;
    }
  }
  resetPaint();
  //emit newObjFile( objFilename );
  
}

void brainApp::sendPaintedVertices( int paintNumber ) {
  vector<int> indices;
  cout << "Total num vertices: " << paintingTexCoord->point.getNum() << endl;
  for (int i=0; i < paintingTexCoord->point.getNum(); i++) {
    if (paintingTexCoord->point[i] == SbVec2f(0, paintNumber)) {
      indices.push_back(i);
    }
  }
  cout << "Num painted vertices: " << indices.size();
  emit paintedVertices( indices );
}

void brainApp::saveSceneGraph() {
  // writes the rootObj scene graph to file.

  QFileDialog* fd = new QFileDialog( (QWidget *) NULL, "Save Scenegraph as:", true );
  fd->setMode( QFileDialog::AnyFile );

  QString fileName;
  if ( fd->exec() == QDialog::Accepted ){
    fileName = fd->selectedFile();

    SoWriteAction wa;
    SoOutput *out = wa.getOutput();
    out->openFile(fileName);
    
    cout << "Saving scene graph to " << fileName << endl;
    wa.apply(this->rootObj);
  }
  delete fd;
}

void brainApp::newFile() {

  QString fn = QFileDialog::getOpenFileName
    ( QString::null, 
      "Brains (*.obj *.obj.gz *.txt *.vertstats)",
      mainWindow );



  if ( !fn.isEmpty() ){
    receivedNewFilename(fn);
  }
  else {
    cerr << "No file loaded" << endl;
  }


}

// slot for when new files are to be loaded    
void brainApp::receivedNewFilename( QString filename ) {
  QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

  if (filename.endsWith(".vertstats") || filename.endsWith(".txt"))
    loadNewVertstatsFile( filename );
  else if (filename.endsWith(".obj") || filename.endsWith(".obj.gz"))
    setOBJFile( filename );
  else 
    cerr << "No file loaded" << endl;
  
  QApplication::restoreOverrideCursor();

}



void brainApp::toggleClipPane() {
  if ( clipSwitch->whichChild.getValue() == SO_SWITCH_ALL ) {
    clipSwitch->whichChild = SO_SWITCH_NONE;
  }
  else {
    clipSwitch->whichChild = SO_SWITCH_ALL;
  }
}



/*! Adds all of the nodes to the scene graph in the correct order
 */
void brainApp::createSceneGraph() {

  rootObj = new SoSeparator;
  rootObj->ref();

  rootObj->addChild( mouseEventCB );

  clipSwitch->addChild( clipper );
  clipSwitch->whichChild = SO_SWITCH_NONE;
  rootObj->addChild( clipSwitch );
  

  cout << "adding texture" << endl;

  // insert texture coordinate
  textureSwitch->addChild( textureComplexity);
  textureSwitch->addChild( texture );
  textureSwitch->addChild( texCoord );
  textureSwitch->addChild( tBind );
  textureSwitch->addChild( textureTrans );

  if (haveTexture == true)
    textureSwitch->whichChild = SO_SWITCH_ALL;
  else
    textureSwitch->whichChild = SO_SWITCH_NONE;

  rootObj->addChild( textureSwitch );
 
  // brain painting stuff
  paintingSwitch->addChild( paintingTexComplexity );
  paintingSwitch->addChild( paintingTexture );
  paintingSwitch->addChild( paintingTexCoord );
  paintingSwitch->addChild( paintingTexBinding );
  paintingSwitch->addChild( paintingTexTrans );
  paintingSwitch->whichChild = SO_SWITCH_NONE;
  rootObj->addChild(paintingSwitch);
  
  cout << "adding geometry" << endl;

  rootObj->addChild(geomDrawStyle);

  rootObj->addChild( geometry );

  // the location sphere
  locationSphereSwitch->whichChild = SO_SWITCH_NONE;
  SoDrawStyle *sphereds = new SoDrawStyle;
  sphereds->style = SoDrawStyleElement::FILLED;
  locationSphereSwitch->addChild( sphereds );

  locationSphereSwitch->addChild( locationSphereMaterial );
  locationSphereSwitch->addChild( locationSphereTranslation );
  locationSphereSwitch->addChild( locationSphere );
  rootObj->addChild(locationSphereSwitch);

  SoGetBoundingBoxAction ba(objViewer->getViewportRegion());
  ba.apply(rootObj);  
  SbBox3f box = ba.getBoundingBox();
  clipper->setValue(box, SbVec3f(1.0f, 0, 0), 1.0f);

  // add the painting code.
  this->addExtselection();
}

void brainApp::flipClipPane() {
  SbVec3f n = clipper->plane.getValue().getNormal();
  n[0] *= -1;
  clipper->plane.setValue(SbPlane(n, -2)); // the -2 is too arbitrary

}

void brainApp::plotFromPickedPoint(float x, float y, float z, int i) {
  plotter->plotVertex(i);
}

void brainApp::traceFromPickedPoint( float x, float y, float z, int i) {
  anatconTracer->traceConnectivity(i);
}

void brainApp::initColourbar() {

  // default min and max
  currentMin = 0;
  currentMax = 1;

  cout << "adding cameras" << endl;
  textureInfo->addChild( camera );
  textureInfo->addChild( new SoDirectionalLight );

  // this array holds the indices into the rectangle which will have 
  // the texture map applied to it.
  SoCoordinate3 *barCoords = new SoCoordinate3;
  textureInfo->addChild(barCoords);  
  barCoords->point.set1Value(0, SbVec3f(-0.9, -1.0, 0));
  barCoords->point.set1Value(1, SbVec3f(0.3, -1.0, 0));
  barCoords->point.set1Value(2, SbVec3f(0.3, 1.0, 0));
  barCoords->point.set1Value(3, SbVec3f(-0.9, 1.0, 0));
  
  SoTextureCoordinate2 *barTex = new SoTextureCoordinate2;
  textureInfo->addChild(barTex);
  barTex->point.set1Value(0, SbVec2f(0,0));
  barTex->point.set1Value(1, SbVec2f(1,0));
  barTex->point.set1Value(2, SbVec2f(1,1));
  barTex->point.set1Value(3, SbVec2f(0,1));
  
  SoFaceSet *barFaceSet = new SoFaceSet;
  barFaceSet->numVertices.set1Value(0,4);
  textureInfo->addChild(barFaceSet);
  
  // The first transform: position the upper text window
  SoTransform *transform = new SoTransform;
  textureInfo->addChild(transform);
  transform->translation.setValue(SbVec3f(.35,.9,0));
  
  // And now add the text
  highColourmapValue = new SoText2;
  highColourmapValue->string = "1";
  textureInfo->addChild(highColourmapValue);
  
  // 3rd quartile text index
  SoTransform *t3q = new SoTransform;
  textureInfo->addChild(t3q);
  t3q->translation.setValue(SbVec3f(0,-.45,0));

  threeQColourmapValue = new SoText2;
  threeQColourmapValue->string = "0.75";
  textureInfo->addChild(threeQColourmapValue);

  // Now the second transform for the lower text value - remember that
  // the previous transform is concatenated to this one
  SoTransform *transform2 = new SoTransform;
  textureInfo->addChild(transform2);
  transform2->translation.setValue(SbVec3f(0,-1.4,0));
  
  // and add the lower text
  lowColourmapValue = new SoText2;
  lowColourmapValue->string = "0";
  textureInfo->addChild(lowColourmapValue);  
  
  cout << "end of colourbar" << endl;

}

void brainApp::plotWindowShow() {
  plotter->show();
}

void brainApp::anatconWindowShow() {
  anatconTracer->show();
}

void brainApp::vinfoPopulate(float f1, float f2, float f3, int vertexNumber) {
  vector<string> headers = vertstats->getDataHeader();
  vector<string>::iterator it;
  char number[20];

  setLocationSphere(vertexNumber);

  vinfoWindow->columnsList->clear();
  
  for( it = headers.begin(); it < headers.end(); it++) {
    vertexColumn tmpColumn = vertstats->getDataColumn((*it).c_str());
    if (vertexNumber != -1)
      sprintf(number, "%f", tmpColumn[vertexNumber]);
    else
      sprintf(number, "%f", 0.0);

    vinfoWindow->columnsList->insertItem
      ( new QListViewItem( vinfoWindow->columnsList, (*it).c_str(),
                           number ));
  }
}


void brainApp::vinfoHeaderSelected() {
  QString text = vinfoWindow->columnsList->selectedItem()->text(0);
  cout << "vinfoHeaderSelected: " << text.latin1() << endl;
  setTextureCoordinates(false, -1, text.latin1());
  updateRangeWidgets();

}

void brainApp::vinfoWindowShow() { 
  vinfoPopulate();
  vinfoWindow->show();
}

// pop up a dialog to set the background colour.
void brainApp::setBGColourDialog() {

  // BUG: for some reason this only works if you select one of the 
  // predefined colours, and not one from the colour wheel. Odd.
  QColor newBGColour = QColorDialog::getColor();
  if (newBGColour.isValid())
    objViewer->setBackgroundColor(SbColor(newBGColour.red(), 
                                                newBGColour.green(), 
                                                newBGColour.blue()));
}
  
void brainApp::addExtselection() {
  

  SoPolygonOffset * po = new SoPolygonOffset;
  po->styles = SoPolygonOffset::FILLED;
  po->factor = 1.0f;
  po->units = 1.0f;
  po->on = TRUE;
  rootObj->insertChild(po, 0);



  selection = new SoExtSelection;
  selection->ref();
  selection->lassoPolicy = SoExtSelection::PART;
  selection->policy = SoSelection::SHIFT;
  selection->lassoType = SoExtSelection::LASSO;
  // #if HAVE_SOEXTSELECTION_LASSOMODE
  selection->lassoMode = SoExtSelection::VISIBLE_SHAPES;
  // #endif // HAVE_SOEXTSELECTION_LASSOMODE

  selection->setLassoColor(SbColor(0.9f,0.7f,0.2f));
  selection->setLassoWidth(3);
  selection->setOverlayLassoPattern(0xf0f0);  // 0011 0011 0011 0011
  selection->animateOverlayLasso(FALSE);
  selection->addStartCallback(start_callback);
  selection->addFinishCallback(finish_callback);
  selection->setTriangleFilterCallback(triangleCB, this);
  selection->setLineSegmentFilterCallback(lineCB, NULL);
  selection->setPointFilterCallback(pointCB, NULL);

  
  SoEventCallback * cb = new SoEventCallback;
  cb->addEventCallback(SoKeyboardEvent::getClassTypeId(),
		       event_cb, this);
  rootObj->insertChild(cb, 0);
  
  selection->addChild(rootObj);
  

  paintRoot = new SoSeparator;
  paintRoot->ref();
  paintRoot->addChild(selection);
  SoSeparator * overlaylines = make_line_sg();
  paintRoot->addChild(overlaylines);

   
  objViewer->setSceneGraph(rootObj);
  //objViewer->setSceneGraph(root);
  objViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  objViewer->setGLRenderAction(new SoBoxHighlightRenderAction(SbViewportRegion()));
  //objViewer->setViewing(FALSE);
   

  SoSearchAction search;
  search.setType(SoExtSelection::getClassTypeId());
  search.setFind(SoSearchAction::TYPE);
  search.setInterest(SoSearchAction::ALL);
  search.apply(rootObj);
  SoPathList & pathlist2 = search.getPaths();


  if(pathlist2.getLength() > 0) {
    (void) fprintf(stderr,
                   "ERROR: You cannot have two ExtSelection nodes beneath each other\n"
                   "in the same scenegraph!\n\n");
   
  }

  // Just check if we can draw a sensible set of overlay lines --
  // multiple cameras makes this very hard, so it's not supported in
  // this example program.

  search.setType(SoCamera::getClassTypeId());
  search.setFind(SoSearchAction::TYPE);
  search.setInterest(SoSearchAction::ALL);
  search.apply(rootObj);
  SoPathList & pathlist = search.getPaths();
  if(pathlist.getLength() > 1) {
    (void) fprintf(stderr,
                   "WARNING: there is more than 1 camera in the loaded\n"
                   "scenegraph -- overlay lines for selection might come\n"
                   "out wrong!\n");

  }


  // Make sure the overlay lines we render are within the same camera
  // scope as the actual geometry from the input file.
  //overlaylines->insertChild(ex1->getCamera(), 0);
  

} 

  

void brainApp::quit() {
  exit(0);
}
